////////////////////////////////////////////////////////////////////////////////////////////////////
// Copyright:      2001/03/01; Andreas Huber <andreas@huber.net>; Zurich; Switzerland
////////////////////////////////////////////////////////////////////////////////////////////////////
// License:        This file is part of the downloadable accompanying material for the article
//                 "Elegant Function Call Wrappers" in the May 2001 issue of the C/C++ Users Journal
//                 (www.cuj.com). The material may be used, copied, distributed, modified and sold
//                 free of charge, provided that this entire notice (copyright, license and
//                 feedback) appears unaltered in all copies of the material.
//                 All material is provided "as is", without express or implied warranty.
////////////////////////////////////////////////////////////////////////////////////////////////////
// Feedback:       Please send comments to andreas@huber.net.
////////////////////////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////////////////////////
// This program shows the usage of DeferCall() in a multithreaded environment.
// The Worker class encapsulates all the platform specific multithreading stuff like events,
// mutexes and so on. The only exception is Sleep(), which is also used in the main program.
////////////////////////////////////////////////////////////////////////////////////////////////////
// CAUTION: For a console program using MFC, a non-standard layout of the main function (_tmain) is
// mandated. To facilitate porting this test program to other platforms I have decided to use the
// standard layout (main). This results in problems during program termination: The debugger reports
// memory leaks and sometimes even deadlocks doing so. Program behaviour inside main() is not
// affected, so just for testing purposes we can safely ignore this.
////////////////////////////////////////////////////////////////////////////////////////////////////



#ifdef WIN32
  // avoid warning "identifier was truncated to '255' characters in the debug information" 
  #pragma warning( disable: 4786 )
#endif



#include "Simple/DeferCall.h"
#include "Worker.h"
#include <iostream>
#include <string>
#include <memory>

#ifdef WIN32
#include <afxmt.h> // Sleep
#endif



////////////////////////////////////////////////////////////////////////////////////////////////////
void OurSleep( unsigned int Milliseconds )
{
  #ifdef WIN32
  Sleep( Milliseconds );
  #endif
}


////////////////////////////////////////////////////////////////////////////////////////////////////
void PrintCharacter( char Character )
{
  OurSleep( 30 ); // We assume printing a character is a really time consuming operation. 
  std::cout << Character;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
void PrintLineImpl( const std::string & Line )
{
  for ( std::string::const_iterator Pos = Line.begin(); Pos != Line.end(); ++Pos )
  {
    PrintCharacter( *Pos );
  }

  PrintCharacter( '\n' );
}

// MFC does not like threads begin created before main() is called. Therefore, the worker is only
// created within main()
static std::auto_ptr< Worker > pPrintWorker;

////////////////////////////////////////////////////////////////////////////////////////////////////
void PrintLine( const std::string & rString )
{
  // Printing is time consuming, we therefore employ another worker to quickly schedule the call
  // and return. So, the other threads have more time at their disposal.
  // Here a non-member function call is deferred, see below for member-function calls.
  pPrintWorker->Do( DeferCall( &PrintLineImpl, rString ) );
}



////////////////////////////////////////////////////////////////////////////////////////////////////
class Bowl
{
  public:
    void Fill( const std::string What )
    {
      CSingleLock( &_Mutex, TRUE );
      _Contents = What;
    }

    const bool IsEmpty() const { return _Contents.empty(); }

    const std::string Empty()
    { 
      CSingleLock( &_Mutex, TRUE );
      std::string Result;
      _Contents.swap( Result );
      return Result;
    }

  private:
    std::string _Contents;
    CCriticalSection _Mutex;
};


////////////////////////////////////////////////////////////////////////////////////////////////////
class Dog
{
  public:
    explicit Dog( const std::string & Name ) : _Name( Name ) {}
    
    enum Volume { Loud, VeryLoud };
    void Bark( Volume HowLoud, unsigned int Times )
    {
      for ( unsigned int Count = 0; Count < Times; ++Count )
      {
        PrintLine( _Name + ": " + ( HowLoud == Loud ? "Bark." : "BARK!" ) );
        OurSleep( 2000 );
      }
    }

    void CheckBowlAndEat( Bowl & MyBowl )
    {
      if ( MyBowl.IsEmpty() )
      {
        PrintLine( _Name + " is hungry but has found his bowl empty." );
        OurSleep( 2000 );
      }
      else
      {
        PrintLine( _Name + " is wolfing down " + MyBowl.Empty() + "." );
        OurSleep( 5000 );
        PrintLine( _Name + " has finished eating." );
      }
    }

  private:
    const std::string _Name;
};


////////////////////////////////////////////////////////////////////////////////////////////////////
class Human
{
  public:
    explicit Human( const std::string & Name ) : _Name( Name ) {}

    void Sleep( double Seconds )
    {
      for ( double Duration = 0.0; Duration < ( Seconds / 4.0 ); ++Duration )
      {
        PrintLine( _Name + ": Zzzzzzzzzzz." );
        OurSleep( 4000 );
      }

      PrintLine( _Name + " wakes up." );
      OurSleep( 4000 );
    }

    void Fill( Bowl & DogsBowl, const std::string & What ) const
    {
      PrintLine( _Name + " fills the bowl with " + What + "." );
      DogsBowl.Fill( What );
      OurSleep( 2000 );
    }

  private:
    const std::string _Name;
};



////////////////////////////////////////////////////////////////////////////////////////////////////
int main(int argc, char* argv[])
{
  Bowl DogsBowl;
  Dog Goofy( "Goofy" );
  Human * pSteve = new Human( "Steve" );

  pPrintWorker = std::auto_ptr< Worker >( new Worker );
  Worker GoofyWorker;
  Worker SteveWorker;

  // Here, non-static member function calls are deferred. Calls to static member functions work
  // similarly: Simply omit the first parameter (the object).
  // Please note how Reference() is used (see article for further details).
  GoofyWorker.Do( DeferCall( &Goofy, &Dog::CheckBowlAndEat, Reference( DogsBowl ) ) );
  GoofyWorker.Do( DeferCall( &Goofy, &Dog::Bark, Dog::Loud, 3 ) );
  GoofyWorker.Do( DeferCall( &Goofy, &Dog::CheckBowlAndEat, Reference( DogsBowl ) ) );
  GoofyWorker.Do( DeferCall( &Goofy, &Dog::Bark, Dog::Loud, 3 ) );
  GoofyWorker.Do( DeferCall( &Goofy, &Dog::CheckBowlAndEat, Reference( DogsBowl ) ) );
  GoofyWorker.Do( DeferCall( &Goofy, &Dog::Bark, Dog::VeryLoud, 5 ) );
  GoofyWorker.Do( DeferCall( &Goofy, &Dog::CheckBowlAndEat, Reference( DogsBowl ) ) );

  SteveWorker.Do( DeferCall( pSteve, &Human::Sleep, 20.0 ) );
  // DeferCall also works with just about any smart pointer. For example, with the following line
  // the ownership of pSteve is transferred to the functor. So, pSteve is deleted as soon as the
  // functor is deleted (inside Worker).
  SteveWorker.Do( DeferCall( std::auto_ptr< Human >( pSteve ),
                             &Human::Fill, Reference( DogsBowl ), std::string( "meat" ) ) );

  // All workers are implicitly destructed just before main() returns.
  // The program does not terminate immediately because Worker::~Worker() waits until the thread
  // has executed all queued calls.
  // Because stack objects are destructed in reverse construction order, the workers are destructed
  // before the other objects. Therefore no access violations occur.
  return 0;
}
