////////////////////////////////////////////////////////////////////////////////////////////////////
// Copyright:      2001/03/01; Andreas Huber <andreas@huber.net>; Zurich; Switzerland
////////////////////////////////////////////////////////////////////////////////////////////////////
// License:        This file is part of the downloadable accompanying material for the article
//                 "Elegant Function Call Wrappers" in the May 2001 issue of the C/C++ Users Journal
//                 (www.cuj.com). The material may be used, copied, distributed, modified and sold
//                 free of charge, provided that this entire notice (copyright, license and
//                 feedback) appears unaltered in all copies of the material.
//                 All material is provided "as is", without express or implied warranty.
////////////////////////////////////////////////////////////////////////////////////////////////////
// Feedback:       Please send comments to andreas@huber.net.
////////////////////////////////////////////////////////////////////////////////////////////////////
#include "Worker.h"
#include <afxwin.h>
#include <afxmt.h>



////////////////////////////////////////////////////////////////////////////////////////////////////
// A CCriticalSection behaves exactly like a mutex. It can be used for in-process thread
// synchronisation only. CMutex, which can also be used across process boundaries, is considerably
// slower.
////////////////////////////////////////////////////////////////////////////////////////////////////
// A CEvent object can be in one of two states: "not set" or "set". A thread locking a "set" CEvent
// object continues executing and the object becomes "not set". A thread locking a "not set"
// object is halted until SetEvent() is called. The thread then continues and the object becomes
// "not set".
////////////////////////////////////////////////////////////////////////////////////////////////////
// CSingleLock - as used throughout this implementation - simply locks the specified resource in the
// constructor and unlocks it in the destructor. This way, we can neither forget to unlock a
// resource nor will we have any problems with exceptions interrupting us when we don't expect it.
// CAUTION: In some places only an unnamed temporary is created, so the lock is destroyed and the
// resource is unlocked before executing the next statement.
////////////////////////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////////////////////////
Worker::Worker() :
  _pThread( NULL ),
  _pCriticalSection( new CCriticalSection ),
  _pThreadHasStarted( new CEvent ),
  _pWork( new CEvent ),
  _pThreadHasTerminated( new CEvent ),
  _IsDestructing( false )
{
  // create a new thread, pass the this pointer to controlling function
  _pThread = AfxBeginThread( &ControllingFunction, this, THREAD_PRIORITY_NORMAL );
  CSingleLock( _pThreadHasStarted, TRUE );
}

////////////////////////////////////////////////////////////////////////////////////////////////////
Worker::~Worker()
{
  _IsDestructing = true;
  _pWork->SetEvent(); // ensure that the thread is not waiting for work

  // ensure the thread has terminated before we continue
  CSingleLock( _pThreadHasTerminated, TRUE );

  // we must not delete the thread, the thread deletes itself automatically 
  delete _pCriticalSection;
  delete _pThreadHasStarted;
  delete _pWork;
  delete _pThreadHasTerminated;
}


////////////////////////////////////////////////////////////////////////////////////////////////////
void Worker::Do(
  Functor * pPieceOfWork
)
{
  CSingleLock Lock( _pCriticalSection, TRUE );

  _Queue.push( pPieceOfWork );
  _pWork->SetEvent();
}


////////////////////////////////////////////////////////////////////////////////////////////////////
void Worker::Work()
{
  while ( !_IsDestructing )
  {
    _pThreadHasStarted->SetEvent();

    CSingleLock( _pWork, TRUE ); // wait until there is something to do

    Functor * pPieceOfWork = GetNextPieceOfWork();

    while ( pPieceOfWork != NULL )
    {
      try
      {
        ( *pPieceOfWork )(); // execute the deferred call
      }
      catch ( ... ) // ensure that no exception is going to interrupt us and terminate the thread
      {
      }

      delete pPieceOfWork;
      pPieceOfWork = GetNextPieceOfWork();
    }
  }

  _pThreadHasTerminated->SetEvent();
}


////////////////////////////////////////////////////////////////////////////////////////////////////
Functor * const Worker::GetNextPieceOfWork()
{
  CSingleLock Lock( _pCriticalSection, TRUE );
  Functor * pPieceOfWork = NULL;

  if ( !_Queue.empty() )
  {
    pPieceOfWork = _Queue.front();
    _Queue.pop();
  }

  return pPieceOfWork;
}


////////////////////////////////////////////////////////////////////////////////////////////////////
unsigned int Worker::ControllingFunction( void * pParam )
{
  static_cast< Worker * >( pParam )->Work();
  return 0;
}