////////////////////////////////////////////////////////////////////////////////////////////////////
// Copyright:      2001/03/01; Andreas Huber <andreas@huber.net>; Zurich; Switzerland
////////////////////////////////////////////////////////////////////////////////////////////////////
// License:        This file is part of the downloadable accompanying material for the article
//                 "Elegant Function Call Wrappers" in the May 2001 issue of the C/C++ Users Journal
//                 (www.cuj.com). The material may be used, copied, distributed, modified and sold
//                 free of charge, provided that this entire notice (copyright, license and
//                 feedback) appears unaltered in all copies of the material.
//                 All material is provided "as is", without express or implied warranty.
////////////////////////////////////////////////////////////////////////////////////////////////////
// Feedback:       Please send comments to andreas@huber.net.
////////////////////////////////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////////////////////////
// This program tests the simple DeferCall() variant, where by-value parameters are passed
// normally while by-reference parameters have to be passed with Reference() (see article).
// The simple variant works perfectly, as long as arguments passed via the Reference() template
// are convertible to the function parameter type without user-defined conversions. So, for
// parameters passed with Reference() no non-copy constructors or conversion operators can be
// invoked when the call is actually executed.
// Whether this variant is sufficient for your project or not depends on the following criteria:
// - Do class or struct type parameters of the asynchronously called functions have non-explicit
//   constructors, which can be invoked providing exactly one argument?
// - Do any of the classes in your project declare non-explicit conversion operators, which convert
//   to any parameter type of the asynchronously called functions?
// If the answer for at least one of the questions is NOT "Definitely not!" you should consider
// employing the "perfect" variant.
////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////////////////////////
// When run, you should see the following output:
// Test0()
// Foo::Bar0()
// Foo::Bar0Const()
// Test1( 1 )
// Foo::Bar1( 1 )
// Foo::Bar1Const( 1 )
// Test2( 1, 2 )
// Foo::Bar2( 1, 2 )
// Foo::Bar2Const( 1, 2 )
// Test3( 1, 2, 3 )
// Foo::Bar3( 1, 2, 3 )
// Foo::Bar3Const( 1, 2, 3 )
// Test4( 1, 2, 3, 4 )
// Foo::Bar4( 1, 2, 3, 4 )
// Foo::Bar4Const( 1, 2, 3, 4 )
// Test5( 1, 2, 3, 4, 5 )
// Foo::Bar5( 1, 2, 3, 4, 5 )
// Foo::Bar5Const( 1, 2, 3, 4, 5 )
// Test6( 1, 2, 3, 4, 5, CopyCount( 2, 0 ) )
// Foo::Bar6( 1, 2, 3, 4, 5, CopyCount( 2, 0 ) )
// Foo::Bar6Const( 1, 2, 3, 4, 5, CopyCount( 0, 0 ) )
////////////////////////////////////////////////////////////////////////////////////////////////////



#ifdef WIN32
  // avoid warning "identifier was truncated to '255' characters in the debug information" 
  #pragma warning( disable: 4786 )
#endif

#include <queue>
#include <iostream>

#include "Simple/DeferCall.h"


using std::cout;
using std::endl;


////////////////////////////////////////////////////////////////////////////////////////////////////
// An object of this class counts how many times it was copied (copy constructor and assignment
// operator).
// It is used to show that the DeferCall mechanism currently copies by-value parameters once more
// than it should. It is possible to get rid of this inefficiency, but only at the cost of
// tayloring the now shared functor subclass templates to the simple and the "perfect" variant. That
// is, the variants would no longer share the functors but would use their own customized set of the
// templates.
// You can get rid of this inefficiency easily without duplication of the functor templates by
// employing partial template specialisation, which MSVC does not support.
////////////////////////////////////////////////////////////////////////////////////////////////////
class CopyCounter
{
  public:
    CopyCounter() :
      CopyCount( 0 ),
      AssignmentCount( 0 )
    {
    }


    CopyCounter( const CopyCounter & Original ) :
      CopyCount( Original.CopyCount + 1 ),
      AssignmentCount( Original.AssignmentCount )
    {
    }

    CopyCounter & operator=( const CopyCounter & Original )
    {
      if ( this != &Original )
      {
        CopyCount = Original.CopyCount;
        AssignmentCount = Original.AssignmentCount + 1;
      }
    }

    unsigned int GetCopyCount() const
    {
      return CopyCount;
    }

    unsigned int GetAssignmentCount() const
    {
      return AssignmentCount;
    }

  private:
    unsigned int CopyCount;
    unsigned int AssignmentCount;
};

template< class OutStream >
OutStream & operator<<( OutStream & Stream, const CopyCounter & Counter )
{
  Stream << "CopyCount( " << Counter.GetCopyCount() << ", " << Counter.GetAssignmentCount() << " )";
  return Stream;
}



void Test0()
{
  cout << "Test0()" << endl;
}

void Test1( unsigned short P1 )
{
  cout << "Test1( " << P1 << " )" << endl;
}

void Test2( unsigned short P1, short P2 )
{
  cout << "Test2( " << P1 << ", " << P2 << " )" << endl;
}

void Test3( unsigned short P1, short P2, const int & P3 )
{
  cout << "Test3( " << P1 << ", " << P2 << ", " << P3 << " )" << endl;
}

void Test4( unsigned short P1, short P2, const int & P3, double & P4 )
{
  cout << "Test4( " << P1 << ", " << P2 << ", " << P3 << ", " << P4 << " )" << endl;
}

void Test5( unsigned short P1, short P2, const int & P3, double & P4, const double & P5 )
{
  cout << "Test5( " << P1 << ", " << P2 << ", " << P3 << ", " << P4 << ", " << P5 << " )" << endl;
}

void Test6( unsigned short P1, short P2, const int & P3, double & P4, const double & P5, const CopyCounter & P6 )
{
  cout << "Test6( " << P1 << ", " << P2 << ", " << P3 << ", " << P4 << ", " << P5 << ", " << P6 << " )" << endl;
}


class Foo
{
  public:
    void Bar0()
    {
      cout << "Foo::Bar0()" << endl;
    }

    void Bar0Const() const
    {
      cout << "Foo::Bar0Const()" << endl;
    }


    void Bar1( unsigned short P1 )
    {
      cout << "Foo::Bar1( " << P1 << " )" << endl;
    }

    void Bar1Const( unsigned short P1 ) const
    {
      cout << "Foo::Bar1Const( " << P1 << " )" << endl;
    }


    void Bar2( unsigned short P1, short P2 )
    {
      cout << "Foo::Bar2( " << P1 << ", " << P2 << " )" << endl;
    }

    void Bar2Const( unsigned short P1, short P2 ) const
    {
      cout << "Foo::Bar2Const( " << P1 << ", " << P2 << " )" << endl;
    }


    void Bar3( unsigned short P1, short P2, const int & P3 )
    {
      cout << "Foo::Bar3( " << P1 << ", " << P2 << ", " << P3 << " )" << endl;
    }

    void Bar3Const( unsigned short P1, short P2, const int & P3 )  const
    {
      cout << "Foo::Bar3Const( " << P1 << ", " << P2 << ", " << P3 << " )" << endl;
    }


    void Bar4( unsigned short P1, short P2, const int & P3, double & P4 )
    {
      cout << "Foo::Bar4( " << P1 << ", " << P2 << ", " << P3 << ", " << P4 << " )" << endl;
    }

    void Bar4Const( unsigned short P1, short P2, const int & P3, double & P4 ) const
    {
      cout << "Foo::Bar4Const( " << P1 << ", " << P2 << ", " << P3 << ", " << P4 << " )" << endl;
    }


    void Bar5( unsigned short P1, short P2, const int & P3, double & P4, const double & P5 )
    {
      cout << "Foo::Bar5( " << P1 << ", " << P2 << ", " << P3 << ", " << P4 << ", " << P5 << " )" << endl;
    }

    void Bar5Const( unsigned short P1, short P2, const int & P3, double & P4, const double & P5 ) const
    {
      cout << "Foo::Bar5Const( " << P1 << ", " << P2 << ", " << P3 << ", " << P4 << ", " << P5 << " )" << endl;
    }

    
    void Bar6( unsigned short P1, short P2, const int & P3, double & P4, const double & P5, const CopyCounter & P6 )
    {
      cout << "Foo::Bar6( " << P1 << ", " << P2 << ", " << P3 << ", " << P4 << ", " << P5 << ", " << P6 << " )" << endl;
    }

    void Bar6Const( unsigned short P1, short P2, const int & P3, double & P4, const double & P5, const CopyCounter & P6 ) const
    {
      cout << "Foo::Bar6Const( " << P1 << ", " << P2 << ", " << P3 << ", " << P4 << ", " << P5 << ", " << P6 << " )" << endl;
    }

};



int main(int argc, char* argv[])
{
  std::queue< Functor * > aQueue;

  Foo aFoo;

  aQueue.push( DeferCall( &Test0 ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar0 ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar0Const ) );

  aQueue.push( DeferCall( &Test1, 1 ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar1, 1 ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar1Const, 1 ) );

  aQueue.push( DeferCall( &Test2, 1, 2 ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar2, 1, 2 ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar2Const, 1, 2 ) );

  aQueue.push( DeferCall( &Test3, 1, 2, 3 ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar3, 1, 2, 3 ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar3Const, 1, 2, 3 ) );

  double Value = 4.0;

  aQueue.push( DeferCall( &Test4, 1, 2, 3, Value ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar4, 1, 2, 3, Value ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar4Const, 1, 2, 3, Value ) );

  aQueue.push( DeferCall( &Test5, 1, 2, 3, Value, Reference( Value ) ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar5, 1, 2, 3, Value, Reference( Value ) ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar5Const, 1, 2, 3, Value, Reference( Value ) ) );

  CopyCounter Counter1;
  CopyCounter Counter2;
  CopyCounter Counter3;

  aQueue.push( DeferCall( &Test6, 1, 2, 3, Value, Reference( Value ), Counter1 ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar6, 1, 2, 3, Value, Reference( Value ), Counter2 ) );
  aQueue.push( DeferCall( &aFoo, &Foo::Bar6Const, 1, 2, 3, Value, Reference( Value ), Reference( Counter3 ) ) );

  ++Value;

  while ( !aQueue.empty() )
  {
    ( *aQueue.front() )();
    delete aQueue.front();
    aQueue.pop();
  }

	return 0;
}
